@@ -4,9 +4,11 @@
 //! contain invalid syntax. In that case the invalid part is skipped.
 
 use crate::{
-    dom::{Cast, KeyNode, NodeSyntax, RootNode},
+    dom::{self, node::DomNode, FromSyntax, Keys, Node},
     syntax::{SyntaxElement, SyntaxKind::*, SyntaxNode, SyntaxToken},
+    util::overlaps,
 };
+use once_cell::unsync::OnceCell;
 use rowan::{GreenNode, NodeOrToken, TextRange};
 use std::{
     cmp,
@@ -16,7 +18,7 @@ use std::{
 };
 
 #[cfg(feature = "serde")]
-use serde_crate::{Deserialize, Serialize};
+use serde::{Deserialize, Serialize};
 
 #[cfg(feature = "schema")]
 use schemars::JsonSchema;
@@ -38,7 +40,6 @@ create_options!(
     /// All the formatting options.
     #[derive(Debug, Clone, Eq, PartialEq)]
     #[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
-    #[cfg_attr(feature = "serde", serde(crate = "serde_crate"))]
     pub struct Options {
         /// Align entries vertically.
         ///
@@ -51,14 +52,22 @@ create_options!(
         /// This applies to comments that are after entries or array items.
         pub align_comments: bool,
 
+        /// If `align_comments` is true, apply the alignment in cases where
+        /// there's only one comment.
+        pub align_single_comments: bool,
+
         /// Put trailing commas for multiline
         /// arrays.
         pub array_trailing_comma: bool,
 
-        /// Automatically expand arrays to multiple lines
-        /// if they're too long.
+        /// Automatically expand arrays to multiple lines once they
+        /// exceed the configured `column_width`.
         pub array_auto_expand: bool,
 
+        /// Expand values (e.g.) inside inline tables
+        /// where possible.
+        pub inline_table_expand: bool,
+
         /// Automatically collapse arrays if they
         /// fit in one line.
         ///
@@ -97,6 +106,9 @@ create_options!(
         /// Alphabetically reorder keys that are not separated by blank lines.
         pub reorder_keys: bool,
 
+        /// Alphabetically reorder array values that are not separated by blank lines.
+        pub reorder_arrays: bool,
+
         /// The maximum amount of consecutive blank lines allowed.
         pub allowed_blank_lines: usize,
 
@@ -110,7 +122,7 @@ pub enum OptionParseError {
     InvalidOption(String),
     InvalidValue {
         key: String,
-        error: Box<dyn std::error::Error>,
+        error: Box<dyn std::error::Error + Send + Sync>,
     },
 }
 
@@ -138,6 +150,7 @@ impl Default for Options {
         Options {
             align_entries: false,
             align_comments: true,
+            align_single_comments: true,
             array_trailing_comma: true,
             array_auto_expand: true,
             array_auto_collapse: true,
@@ -147,10 +160,12 @@ impl Default for Options {
             column_width: 80,
             indent_tables: false,
             indent_entries: false,
+            inline_table_expand: true,
             trailing_newline: true,
             allowed_blank_lines: 2,
             indent_string: "  ".into(),
             reorder_keys: false,
+            reorder_arrays: false,
             crlf: false,
         }
     }
@@ -159,24 +174,40 @@ impl Default for Options {
 impl Options {
     fn newline(&self) -> &'static str {
         if self.crlf {
-            "\r\n".into()
+            "\r\n"
         } else {
-            "\n".into()
+            "\n"
         }
     }
 
     fn newlines(&self, count: usize) -> impl Iterator<Item = &'static str> {
         repeat(self.newline()).take(usize::min(count, self.allowed_blank_lines + 1))
     }
+
+    fn should_align_comments(&self, comment_count: usize) -> bool {
+        (comment_count != 1 || self.align_single_comments) && self.align_comments
+    }
 }
 
-#[derive(Debug, Clone, Default)]
+#[derive(Debug, Clone)]
 struct Context {
     indent_level: usize,
     force_multiline: bool,
+    errors: Rc<[TextRange]>,
     scopes: Rc<ScopedOptions>,
 }
 
+impl Default for Context {
+    fn default() -> Self {
+        Self {
+            indent_level: Default::default(),
+            force_multiline: Default::default(),
+            errors: Rc::from([]),
+            scopes: Default::default(),
+        }
+    }
+}
+
 impl Context {
     /// Update options based on the text range.
     fn update_options(&self, opts: &mut Options, range: TextRange) {
@@ -187,6 +218,16 @@ impl Context {
         }
     }
 
+    fn error_at(&self, range: TextRange) -> bool {
+        for error_range in self.errors.iter().copied() {
+            if overlaps(range, error_range) {
+                return true;
+            }
+        }
+
+        false
+    }
+
     fn indent<'o>(&self, opts: &'o Options) -> impl Iterator<Item = &'o str> {
         repeat(opts.indent_string.as_ref()).take(self.indent_level)
     }
@@ -197,14 +238,21 @@ pub fn format_green(green: GreenNode, options: Options) -> String {
     format_syntax(SyntaxNode::new_root(green), options)
 }
 
-/// Parses then formats a TOML document, ignoring errors.
+/// Parses then formats a TOML document, skipping ranges that contain syntax errors.
 pub fn format(src: &str, options: Options) -> String {
-    format_syntax(crate::parser::parse(src).into_syntax(), options)
+    let p = crate::parser::parse(src);
+
+    let ctx = Context {
+        errors: p.errors.iter().map(|err| err.range).collect(),
+        ..Context::default()
+    };
+
+    format_impl(p.into_syntax(), options, ctx)
 }
 
 /// Formats a parsed TOML syntax tree.
 pub fn format_syntax(node: SyntaxNode, options: Options) -> String {
-    let mut s = format_impl(node, options.clone(), Context::default()).to_string();
+    let mut s = format_impl(node, options.clone(), Context::default());
 
     s = s.trim_end().into();
 
@@ -218,16 +266,23 @@ pub fn format_syntax(node: SyntaxNode, options: Options) -> String {
 /// Formats a DOM root node with given scopes.
 ///
 /// **This doesn't check errors of the DOM.**
-pub fn format_with_scopes(dom: RootNode, options: Options, scopes: ScopedOptions) -> String {
-    let mut c = Context::default();
-    c.scopes = Rc::new(scopes);
+pub fn format_with_scopes(
+    dom: Node,
+    options: Options,
+    errors: &[TextRange],
+    scopes: ScopedOptions,
+) -> String {
+    let c = Context {
+        scopes: Rc::new(scopes),
+        errors: errors.into(),
+        ..Context::default()
+    };
 
     let mut s = format_impl(
-        dom.syntax().into_node().unwrap(),
+        dom.syntax().unwrap().clone().into_node().unwrap(),
         options.clone(),
-        Context::default(),
-    )
-    .to_string();
+        c,
+    );
 
     s = s.trim_end().into();
 
@@ -240,31 +295,40 @@ pub fn format_with_scopes(dom: RootNode, options: Options, scopes: ScopedOptions
 
 /// Formats a DOM root node with given scopes.
 ///
-/// All the scope keys must be valid glob patterns,
-/// otherwise this function will panic!
-///
 /// **This doesn't check errors of the DOM.**
-pub fn format_with_path_scopes<I: IntoIterator<Item = (String, OptionsIncomplete)>>(
-    dom: RootNode,
+pub fn format_with_path_scopes<I, S>(
+    dom: Node,
     options: Options,
+    errors: &[TextRange],
     scopes: I,
-) -> String {
-    let mut c = Context::default();
+) -> Result<String, dom::Error>
+where
+    I: IntoIterator<Item = (S, OptionsIncomplete)>,
+    S: AsRef<str>,
+{
+    let mut c = Context {
+        errors: errors.into(),
+        ..Context::default()
+    };
 
     let mut s = Vec::new();
 
     for (scope, opts) in scopes {
-        let pat = glob::Pattern::new(&scope).unwrap();
-        for (p2, node) in dom.iter() {
-            if pat.matches(&p2.dotted()) {
-                s.extend(node.text_ranges().into_iter().map(|r| (r, opts.clone())))
-            }
+        let keys: Keys = scope.as_ref().parse()?;
+        let matched = dom.find_all_matches(keys, false)?;
+
+        for (_, node) in matched {
+            s.extend(node.text_ranges().into_iter().map(|r| (r, opts.clone())));
         }
     }
 
     c.scopes = Rc::new(ScopedOptions::from_iter(s));
 
-    let mut s = format_impl(dom.syntax().into_node().unwrap(), options.clone(), c).to_string();
+    let mut s = format_impl(
+        dom.syntax().unwrap().clone().into_node().unwrap(),
+        options.clone(),
+        c,
+    );
 
     s = s.trim_end().into();
 
@@ -272,7 +336,7 @@ pub fn format_with_path_scopes<I: IntoIterator<Item = (String, OptionsIncomplete
         s += options.newline();
     }
 
-    s
+    Ok(s)
 }
 
 fn format_impl(node: SyntaxNode, options: Options, context: Context) -> String {
@@ -295,16 +359,28 @@ fn format_impl(node: SyntaxNode, options: Options, context: Context) -> String {
 struct FormattedEntry {
     syntax: SyntaxElement,
     key: String,
+    /// This field is used to cache the "cleaned" version of the key and should only
+    /// be accessed through the `cleaned_key` helpers method.
+    cleaned_key: OnceCell<Vec<String>>,
     value: String,
     comment: Option<String>,
 }
 
+impl FormattedEntry {
+    fn cleaned_key(&self) -> &Vec<String> {
+        self.cleaned_key.get_or_init(|| {
+            self.key
+                .replace(['\'', '"'], "")
+                .split('.')
+                .map(ToOwned::to_owned)
+                .collect()
+        })
+    }
+}
+
 impl PartialEq for FormattedEntry {
     fn eq(&self, other: &Self) -> bool {
-        self.key
-            .replace('\'', "")
-            .replace('"', "")
-            .eq(&other.key.replace('\'', "").replace('"', ""))
+        self.cleaned_key().eq(other.cleaned_key())
     }
 }
 
@@ -312,19 +388,13 @@ impl Eq for FormattedEntry {}
 
 impl PartialOrd for FormattedEntry {
     fn partial_cmp(&self, other: &Self) -> Option<cmp::Ordering> {
-        self.key
-            .replace('\'', "")
-            .replace('"', "")
-            .partial_cmp(&other.key.replace('\'', "").replace('"', ""))
+        Some(self.cmp(other))
     }
 }
 
 impl Ord for FormattedEntry {
     fn cmp(&self, other: &Self) -> cmp::Ordering {
-        self.key
-            .replace('\'', "")
-            .replace('"', "")
-            .cmp(&other.key.replace('\'', "").replace('"', ""))
+        self.cleaned_key().cmp(other.cleaned_key())
     }
 }
 
@@ -366,8 +436,7 @@ fn format_root(node: SyntaxNode, options: &Options, context: &Context) -> String
     let mut context = context.clone();
 
     // Table key for determining indents
-    let mut last_table_key = None;
-    let mut table_key_indent_history: Vec<(KeyNode, usize)> = Vec::new();
+    let mut table_key_indent_history: Vec<(Keys, usize)> = Vec::new();
 
     fn add_comments(
         comments: &mut Vec<String>,
@@ -379,9 +448,9 @@ fn format_root(node: SyntaxNode, options: &Options, context: &Context) -> String
 
         for (idx, comment) in comments.drain(0..).enumerate() {
             if idx != 0 {
-                *formatted += &options.newline();
+                *formatted += options.newline();
             }
-            formatted.extend(context.indent(&options));
+            formatted.extend(context.indent(options));
             *formatted += &comment;
         }
 
@@ -389,45 +458,46 @@ fn format_root(node: SyntaxNode, options: &Options, context: &Context) -> String
     }
 
     let mut dangling_newline_count = 0;
+    let mut scoped_options = options.clone();
 
     for c in node.children_with_tokens() {
-        let mut options = options.clone();
-        context.update_options(&mut options, c.text_range());
+        if context.error_at(c.text_range()) {
+            formatted += &c.to_string();
+            continue;
+        }
+
+        let c_range = c.text_range();
 
         match c {
             NodeOrToken::Node(node) => match node.kind() {
                 TABLE_ARRAY_HEADER | TABLE_HEADER => {
-                    if add_entries(&mut entry_group, &mut formatted, &options, &context) {
-                        formatted += &options.newline();
+                    if add_entries(&mut entry_group, &mut formatted, &scoped_options, &context) {
+                        formatted += scoped_options.newline();
                         skip_newlines = 0;
                     }
 
+                    scoped_options = options.clone();
+                    context.update_options(&mut scoped_options, c_range);
+
                     // We treat everything as indented other than table headers from now on.
-                    if options.indent_entries && context.indent_level == 0 {
+                    if scoped_options.indent_entries && context.indent_level == 0 {
                         context.indent_level = 1;
                     }
 
-                    if let Some(key) = node.first_child().map(Into::into).and_then(KeyNode::cast) {
-                        if let Some(last_key) = last_table_key {
-                            if options.indent_tables {
-                                context.indent_level = table_indent_level(
-                                    &table_key_indent_history,
-                                    &last_key,
-                                    &key,
-                                    context.indent_level,
-                                    if options.indent_entries { 1 } else { 0 },
-                                );
-                            }
+                    if let Some(key) = node.first_child().map(Into::into).map(Keys::from_syntax) {
+                        if scoped_options.indent_tables {
+                            context.indent_level = table_indent_level(
+                                &table_key_indent_history,
+                                &key,
+                                if scoped_options.indent_entries { 1 } else { 0 },
+                            );
                         }
-
                         table_key_indent_history.push((key.clone(), context.indent_level));
-
-                        last_table_key = Some(key);
                     }
 
                     let mut header_context = context.clone();
 
-                    if options.indent_entries {
+                    if scoped_options.indent_entries {
                         header_context.indent_level = header_context.indent_level.saturating_sub(1);
                     }
 
@@ -435,32 +505,40 @@ fn format_root(node: SyntaxNode, options: &Options, context: &Context) -> String
                         &mut comment_group,
                         &mut formatted,
                         &header_context,
-                        &options,
+                        &scoped_options,
                     ) {
-                        formatted += &options.newline();
+                        formatted += scoped_options.newline();
                         skip_newlines = 0;
                     }
 
-                    let header = format_table_header(node, &options, &header_context);
+                    let header = format_table_header(node, &scoped_options, &header_context);
                     let comment = header.trailing_comment();
 
-                    if options.indent_tables {
-                        formatted.extend(header_context.indent(&options));
+                    if scoped_options.indent_tables {
+                        formatted.extend(header_context.indent(&scoped_options));
                     }
 
-                    header.write_to(&mut formatted, &options);
+                    header.write_to(&mut formatted, &scoped_options);
                     if let Some(c) = comment {
                         formatted += " ";
                         formatted += &c;
                     }
                 }
                 ENTRY => {
-                    if add_comments(&mut comment_group, &mut formatted, &context, &options) {
-                        formatted += &options.newline();
+                    scoped_options = options.clone();
+                    context.update_options(&mut scoped_options, c_range);
+
+                    if add_comments(
+                        &mut comment_group,
+                        &mut formatted,
+                        &context,
+                        &scoped_options,
+                    ) {
+                        formatted += scoped_options.newline();
                         skip_newlines = 0;
                     }
 
-                    entry_group.push(format_entry(node, &options, &context));
+                    entry_group.push(format_entry(node, &scoped_options, &context));
                     skip_newlines += 1;
                 }
                 _ => unreachable!(),
@@ -481,63 +559,62 @@ fn format_root(node: SyntaxNode, options: &Options, context: &Context) -> String
                     }
 
                     if newline_count > 1 {
-                        add_comments(&mut comment_group, &mut formatted, &context, &options);
-                        add_entries(&mut entry_group, &mut formatted, &options, &context);
+                        add_comments(
+                            &mut comment_group,
+                            &mut formatted,
+                            &context,
+                            &scoped_options,
+                        );
+                        add_entries(&mut entry_group, &mut formatted, &scoped_options, &context);
                         skip_newlines = 0;
                     }
 
-                    formatted.extend(options.newlines(newline_count.saturating_sub(skip_newlines)));
+                    formatted.extend(
+                        scoped_options.newlines(newline_count.saturating_sub(skip_newlines)),
+                    );
                 }
                 COMMENT => {
-                    if add_entries(&mut entry_group, &mut formatted, &options, &context) {
-                        formatted += &options.newline();
+                    if add_entries(&mut entry_group, &mut formatted, &scoped_options, &context) {
+                        formatted += scoped_options.newline();
                         skip_newlines = 0;
                     }
                     comment_group.push(token.text().to_string());
                     skip_newlines += 1;
                 }
                 WHITESPACE => {}
-                _ => unreachable!(),
+                _ => formatted += token.text(),
             },
         }
     }
 
-    add_comments(&mut comment_group, &mut formatted, &context, options);
-    add_entries(&mut entry_group, &mut formatted, options, &context);
+    add_comments(
+        &mut comment_group,
+        &mut formatted,
+        &context,
+        &scoped_options,
+    );
+    add_entries(&mut entry_group, &mut formatted, &scoped_options, &context);
 
     formatted
 }
 
-/// Determine the indentation level based on 2 consecutive table keys.
+/// Determine the indentation level using the indentation history.
+///
+/// The latest key that is a strict prefix is used and indented. If none is found, the default
+/// indentation is used.
 fn table_indent_level(
-    history: &[(KeyNode, usize)],
-    previous_key: &KeyNode,
-    current_key: &KeyNode,
-    indent: usize,
+    history: &[(Keys, usize)],
+    current_key: &Keys,
     default_indent: usize,
 ) -> usize {
-    if previous_key == current_key {
-        return indent;
-    }
-
-    if previous_key.common_prefix_count(current_key) > 0 && current_key.contains(previous_key) {
-        return indent + 1;
-    } else if previous_key.common_prefix_count(current_key) > 0
-        && current_key.key_count() >= previous_key.key_count()
-    {
-        return indent;
-    }
-
-    if !previous_key.is_part_of(current_key) && !previous_key.contains(current_key) {
-        return default_indent;
-    }
-
-    // Find a table with the same name, likely to be on the same level.
-    if let Some((_, indent)) = history.iter().rev().find(|(key, _)| key == current_key) {
-        return *indent;
-    }
-
-    indent.saturating_sub(1)
+    history
+        .iter()
+        .rev()
+        .find_map(|(previous_key, indent)| {
+            (current_key.contains(previous_key) && current_key != previous_key)
+                .then_some(*indent + 1)
+        })
+        .unwrap_or(default_indent)
 }
 
 /// Add entries to the formatted string.
@@ -567,11 +644,11 @@ fn add_entries(
                 )
                 .unwrap_or(0);
 
-            let line_count = entry.value.split("\n").count();
+            let line_count = entry.value.split('\n').count();
 
             // check each line of the value
             // for the first line we include the actual indent, key, and the eq parts as well
-            for (idx, line) in entry.value.split("\n").enumerate() {
+            for (idx, line) in entry.value.split('\n').enumerate() {
                 let mut chars_count = line.chars().count();
                 if idx == 0 {
                     chars_count += indent_chars_count;
@@ -603,7 +680,11 @@ fn add_entries(
 
                     entry.value.clear();
 
-                    entry.comment = value.trailing_comment();
+                    if let Some(c) = value.trailing_comment() {
+                        debug_assert!(entry.comment.is_none() || entry.comment.clone().unwrap() == c);
+                        entry.comment = Some(c);
+                    }
+
                     value.write_to(&mut entry.value, options);
                     break;
                 }
@@ -611,6 +692,7 @@ fn add_entries(
         }
     }
 
+    let mut comment_count = 0;
     // Transform the entries into generic rows that can be aligned.
     let rows = entry_group
         .drain(0..)
@@ -623,18 +705,20 @@ fn add_entries(
             row.push(e.value);
             if let Some(c) = e.comment {
                 row.push(c);
+                comment_count += 1;
             }
 
             row
         })
         .collect::<Vec<_>>();
 
+    let align_comments = options.should_align_comments(comment_count);
     *formatted += &format_rows(
-        if !options.align_entries && !options.align_comments {
+        if !options.align_entries && !align_comments {
             0..0
-        } else if !options.align_entries && options.align_comments {
+        } else if !options.align_entries && align_comments {
             3..usize::MAX
-        } else if options.align_entries && !options.align_comments {
+        } else if options.align_entries && !align_comments {
             0..3
         } else {
             0..usize::MAX
@@ -688,6 +772,7 @@ fn format_entry(node: SyntaxNode, options: &Options, context: &Context) -> Forma
     FormattedEntry {
         syntax: node.into(),
         key,
+        cleaned_key: OnceCell::new(),
         value,
         comment,
     }
@@ -701,7 +786,7 @@ fn format_key(node: SyntaxNode, formatted: &mut String, _options: &Options, _con
             NodeOrToken::Token(t) => match t.kind() {
                 WHITESPACE | NEWLINE => {}
                 _ => {
-                    *formatted += &t.text();
+                    *formatted += t.text();
                 }
             },
         }
@@ -767,6 +852,12 @@ fn format_inline_table(
     let mut formatted = String::new();
     let mut comment = None;
 
+    let mut context = context.clone();
+    if context.force_multiline {
+        context.force_multiline = options.inline_table_expand;
+    }
+    let context = &context;
+
     let child_count = node.children().count();
 
     if node.children().count() == 0 {
@@ -815,7 +906,7 @@ fn format_inline_table(
                     debug_assert!(comment.is_none());
                     comment = Some(t.text().into());
                 }
-                _ => unreachable!(),
+                _ => formatted += t.text(),
             },
         }
     }
@@ -836,7 +927,6 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
     let mut multiline = is_array_multiline(&node) || context.force_multiline;
 
     let mut formatted = String::new();
-    let mut trailing_comment = None;
 
     // We always try to collapse it if possible.
     if can_collapse_array(&node) && options.array_auto_collapse && !context.force_multiline {
@@ -846,16 +936,28 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
     // We use the same strategy as for entries, refer to [`format_root`].
     let mut skip_newlines = 0;
 
-    // Formatted value, and optional trailing comment.
-    // The value should also include the comma at the end if needed.
+    // Formatted value, optional trailing comment
+    // The value must not include the comma at the end.
     let mut value_group: Vec<(String, Option<String>)> = Vec::new();
+    let mut commas_group: Vec<bool> = Vec::new();
 
     let add_values = |value_group: &mut Vec<(String, Option<String>)>,
+                      commas_group: &mut Vec<bool>,
                       formatted: &mut String,
                       context: &Context|
      -> bool {
         let were_values = !value_group.is_empty();
 
+        if options.reorder_arrays {
+            value_group.sort_unstable_by(|x, y| x.0.cmp(&y.0));
+        }
+
+        for (has_comma, p) in commas_group.drain(0..).zip(value_group.iter_mut()) {
+            if has_comma {
+                p.0 += ","
+            };
+        }
+
         if !multiline {
             for (idx, (val, comment)) in value_group.drain(0..).enumerate() {
                 debug_assert!(comment.is_none());
@@ -869,6 +971,7 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
             return were_values;
         }
 
+        let mut comment_count = 0;
         let rows = value_group
             .drain(0..)
             .map(|(value, comment)| {
@@ -878,18 +981,16 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
                 row.push(value);
                 if let Some(c) = comment {
                     row.push(c);
+                    comment_count += 1;
                 }
 
                 row
             })
             .collect::<Vec<_>>();
 
+        let align_comments = options.should_align_comments(comment_count);
         *formatted += &format_rows(
-            if options.align_comments {
-                0..usize::MAX
-            } else {
-                0..0
-            },
+            if align_comments { 0..usize::MAX } else { 0..0 },
             1..usize::MAX,
             &rows,
             options.newline(),
@@ -914,7 +1015,7 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
         match c {
             NodeOrToken::Node(n) => match n.kind() {
                 VALUE => {
-                    if multiline && formatted.ends_with("[") {
+                    if multiline && formatted.ends_with('[') {
                         formatted += options.newline();
                     }
 
@@ -923,16 +1024,20 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
 
                     val.write_to(&mut val_string, options);
 
-                    if node_index < node_count - 1 || (multiline && options.array_trailing_comma) {
-                        val_string += ",";
-                    }
+                    let has_comma =
+                        node_index < node_count - 1 || (multiline && options.array_trailing_comma);
+                    commas_group.push(has_comma);
 
                     value_group.push((val_string, val.trailing_comment()));
                     skip_newlines += 1;
 
                     node_index += 1;
                 }
-                _ => unreachable!(),
+                _ => {
+                    if cfg!(debug_assertions) {
+                        unreachable!()
+                    }
+                }
             },
             NodeOrToken::Token(t) => match t.kind() {
                 BRACKET_START => {
@@ -942,7 +1047,12 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
                     }
                 }
                 BRACKET_END => {
-                    add_values(&mut value_group, &mut formatted, &inner_context);
+                    add_values(
+                        &mut value_group,
+                        &mut commas_group,
+                        &mut formatted,
+                        &inner_context,
+                    );
 
                     if multiline {
                         if !formatted.ends_with('\n') {
@@ -974,21 +1084,22 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
                     }
 
                     if newline_count > 1 {
-                        add_values(&mut value_group, &mut formatted, &inner_context);
+                        add_values(
+                            &mut value_group,
+                            &mut commas_group,
+                            &mut formatted,
+                            &inner_context,
+                        );
                         skip_newlines = 0;
                     }
 
-                    if !formatted.ends_with("]") {
-                        formatted
-                            .extend(options.newlines(newline_count.saturating_sub(skip_newlines)));
-                    }
+                    formatted.extend(options.newlines(newline_count.saturating_sub(skip_newlines)));
                 }
                 COMMENT => {
                     let newline_before = t
                         .siblings_with_tokens(rowan::Direction::Prev)
-                        .skip(1) // skip is needed because the iterator includes the actual token
-                        .filter(|s| s.kind() != WHITESPACE)
-                        .next()
+                        .skip(1)
+                        .find(|s| s.kind() != WHITESPACE)
                         .map(|s| s.kind() == NEWLINE)
                         .unwrap_or(false);
 
@@ -998,14 +1109,17 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
                         continue;
                     }
 
-                    if add_values(&mut value_group, &mut formatted, &inner_context) {
+                    if add_values(
+                        &mut value_group,
+                        &mut commas_group,
+                        &mut formatted,
+                        &inner_context,
+                    ) {
                         formatted += options.newline();
                         skip_newlines = 0;
                     }
 
-                    if formatted.ends_with(']') {
-                        trailing_comment = Some(t.text().into());
-                    } else if formatted.ends_with('[') {
+                    if formatted.ends_with('[') {
                         formatted += " ";
                         formatted += t.text();
                     } else {
@@ -1018,11 +1132,11 @@ fn format_array(node: SyntaxNode, options: &Options, context: &Context) -> impl
         }
     }
 
-    if formatted.len() == 0 {
+    if formatted.is_empty() {
         formatted = "[]".into();
     }
 
-    (node.into(), formatted, trailing_comment)
+    (node.into(), formatted, None)
 }
 
 fn format_table_header(
@@ -1036,7 +1150,7 @@ fn format_table_header(
     for c in node.children_with_tokens() {
         match c {
             NodeOrToken::Node(n) => {
-                format_key(n, &mut &mut formatted, options, context);
+                format_key(n, &mut formatted, options, context);
             }
             NodeOrToken::Token(t) => match t.kind() {
                 BRACKET_START | BRACKET_END => formatted += t.text(),
@@ -1045,7 +1159,7 @@ fn format_table_header(
                     debug_assert!(comment.is_none());
                     comment = Some(t.text().to_string());
                 }
-                _ => unreachable!(),
+                _ => formatted += t.text(),
             },
         }
     }
@@ -1070,6 +1184,7 @@ impl<T: AsRef<str>> FormattedItem for (SyntaxElement, T, Option<T>) {
 
 trait FormattedItem {
     fn syntax(&self) -> SyntaxElement;
+    #[allow(clippy::ptr_arg)]
     fn write_to(&self, formatted: &mut String, options: &Options);
     fn trailing_comment(&self) -> Option<String>;
 }
@@ -1085,7 +1200,7 @@ impl NewlineCount for &str {
 }
 
 // FIXME(docs)
-fn format_rows<'r, R, S>(
+fn format_rows<R, S>(
     align_range: Range<usize>,
     separator_range: Range<usize>,
     rows: &[R],
@@ -1101,9 +1216,8 @@ where
     // We currently don't support vertical alignment of complex data.
     let can_align = rows
         .iter()
-        .map(|r| r.as_ref().iter())
-        .flatten()
-        .all(|s| !s.as_ref().contains("\n"));
+        .flat_map(|r| r.as_ref().iter())
+        .all(|s| !s.as_ref().contains('\n'));
 
     let diff_widths = |range: Range<usize>, row: &R| -> usize {
         let mut max_width = 0_usize;
